/*
 * This code is partially taken from com.atlassian.confluence.plugins.macros.dashboard.SpacesListMacro.
 * See below for details.
 */
package de.communardo.confluence.plugins.displaysubspace;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.collections.comparators.ReverseComparator;
import org.apache.commons.lang.StringUtils;
import org.joda.time.DateMidnight;
import org.joda.time.base.AbstractInstant;

import com.atlassian.bandana.BandanaManager;
import com.atlassian.confluence.labels.Label;
import com.atlassian.confluence.labels.LabelManager;
import com.atlassian.confluence.languages.LocaleManager;
import com.atlassian.confluence.renderer.PageContext;
import com.atlassian.confluence.renderer.radeox.macros.MacroUtils;
import com.atlassian.confluence.security.Permission;
import com.atlassian.confluence.setup.bandana.ConfluenceBandanaContext;
import com.atlassian.confluence.setup.settings.Settings;
import com.atlassian.confluence.setup.settings.SettingsManager;
import com.atlassian.confluence.spaces.Space;
import com.atlassian.confluence.spaces.SpaceGroup;
import com.atlassian.confluence.spaces.SpaceGroupManager;
import com.atlassian.confluence.spaces.SpaceManager;
import com.atlassian.confluence.spaces.SpaceType;
import com.atlassian.confluence.spaces.SpacesQuery;
import com.atlassian.confluence.spaces.SpacesQuery.Builder;
import com.atlassian.confluence.user.AuthenticatedUserThreadLocal;
import com.atlassian.confluence.user.UserInterfaceState;
import com.atlassian.confluence.util.GeneralUtil;
import com.atlassian.confluence.util.SpaceComparator;
import com.atlassian.confluence.util.i18n.I18NBean;
import com.atlassian.renderer.RenderContext;
import com.atlassian.renderer.v2.RenderMode;
import com.atlassian.renderer.v2.macro.BaseMacro;
import com.atlassian.renderer.v2.macro.MacroException;
import com.atlassian.user.User;
import com.atlassian.util.profiling.UtilTimerStack;
import com.opensymphony.util.TextUtils;

import de.communardo.confluence.plugins.addlink.v2.Link;
import de.communardo.confluence.plugins.addlink.v2.LinkTreeService;
import de.communardo.confluence.plugins.util.ComponentManager;

/**
 * Renders a a subspace navigation based on the given parameters.
 * 
 * @author Tino winkler &lt;tino.winkler@communardo.de&gt;
 * 
 */
public class SubSpacesListMacro extends BaseMacro {

    public static final String PARAMETER_CURRENTSPACE = "currentSpace";
    private static final String PARAMETER_MAXDEPTH = "maxDepth";
    public final static String MACRO_NAME = "subspaces";
    private static final String PARAM_CSSCLASS = "cssclass";
    public static final String PARAM_SHOW_SPACE_ICONS = "showSpaceIcons";
    private static final String PARAM_SHOW_HOME_BUTTON = "showHomeButton";
    private static final String CSS_STYLE_CLASS = "css_style_class";
    private final static String HIDE_SPACES = "hideSpaces";
    private final static String ADDLINK = "addlink";
    private final static String SORT = "sort";
    private final static String STYLE = "style";
    private final static String STYLE_MENU = "menue";
    private final static String PARAMETER_SHOWROOT = "showRoot";
    private final static String STYLE_PLAIN = "plain";
    private final static String STYLE_LIST = "list";
    private final static String STYLE_CLASS_MENU = "subspaces-menu";
    private final static String STYLE_CLASS_PLAIN = "subspaces-plain";
    private final static String KEY = "key";
    private final static String DISPLAY_EMPTY_MESSAGE = "displayEmptyMessage";
    private static final String STYLE_CLASS_LIST = "subspaces-list";

    private BandanaManager bandanaManager;
    private LocaleManager localeManager;
    private SettingsManager settingsManager;
    private LabelManager labelManager;
    private SpaceManager spaceManager;
    private DashboardMacroSupport dashboardMacroSupport;
    private SpaceGroupManager spaceGroupManager;
    private LinkTreeService linkTreeService;

    /**
     * Entry point for macro rendering.
     * <p>
     * The original method has been completely rewritten.
     * </p>
     * 
     * @param macroParameter
     * @return
     * @throws IllegalArgumentException
     * @throws IOException
     */
    public String execute(Map macroParameter, String body, RenderContext renderContext) throws MacroException {

        // WARNING: This macros implementation is tightly linked to the RecentlyUpdatedContentMacro,
        // used by the dashboard to display the recently updated content of the spaces listed by
        // this macro. Some of the common processing has been extracted into the
        // DashboardMacroSupport class, but not all.
        UtilTimerStack.push("Macro: {subspaces}");
        try {
            Map<String, Object> contextMap = getDefaultContextMap(macroParameter, renderContext);

            String sort = (String) macroParameter.get(SORT);
            String style = (String) macroParameter.get(STYLE);
            String hierarchyDepth = (String) macroParameter.get(PARAMETER_MAXDEPTH);

            int hierarchyDepthInteger = StringUtils.isNotEmpty(hierarchyDepth)
                    && StringUtils.isNumeric(hierarchyDepth.trim()) ? new Integer(hierarchyDepth.trim()) : -1;

            // define the style class
            contextMap.put(CSS_STYLE_CLASS, getStyleClass(style, (String) macroParameter.get(PARAM_CSSCLASS)));

            Space parentSpace = ComponentManager.getSpaceManager().getSpace((String) contextMap.get(KEY));

            User user = AuthenticatedUserThreadLocal.getUser();
            // get the permitted Spaces... possible reason for performance las... use this method
            // carefully

            String paramHideSpaces = (String) macroParameter.get(HIDE_SPACES);
            boolean hideSpaces = false;
            if (paramHideSpaces != null && Boolean.parseBoolean(paramHideSpaces)) {
                hideSpaces = true;
            }
            contextMap.put(HIDE_SPACES, hideSpaces);

            List<Space> permittedSpaces = new ArrayList<Space>();
            if (hideSpaces == false || (hideSpaces == true && StringUtils.equalsIgnoreCase(style, STYLE_LIST))) {

                Builder newQuery = SpacesQuery.newQuery();
                newQuery.forUser(user);
                newQuery.withSpaceType(SpaceType.GLOBAL);
                permittedSpaces = ComponentManager.getSpaceManager().getAllSpaces(newQuery.build());
            }

            if (StringUtils.equalsIgnoreCase(style, STYLE_LIST)) {
                String displaySingleTab = (String) macroParameter.get("0");
                prepareListContextMap(displaySingleTab, sort, contextMap, hierarchyDepthInteger, parentSpace,
                        permittedSpaces);
                return HierarchyRenderer.render(contextMap,
                        "communardo_plugins/displaysubspace/displayspacehierarchie_tree/macro-subspacelist.vm",
                        "communardo_plugins/displaysubspace/displayspacehierarchie_tree/treeNode.vm",
                        (String) contextMap.get(PARAMETER_CURRENTSPACE), permittedSpaces);
            } else {
                prepareDefaultContextMap(sort, contextMap, hierarchyDepthInteger, parentSpace, permittedSpaces);
                return HierarchyRenderer.render(contextMap,
                        "communardo_plugins/displaysubspace/displayspacehierarchy/subspacelist.vm",
                        "communardo_plugins/displaysubspace/displayspacehierarchy/treeNode.vm",
                        (String) contextMap.get(PARAMETER_CURRENTSPACE), permittedSpaces);
            }
        } finally {
            UtilTimerStack.pop("Macro: {subspaces}");
        }
    }

    private String getStyleClass(String style, String optionalCssClass) {

        if (optionalCssClass != null && optionalCssClass.length() > 0) {
            return optionalCssClass;
        } else if (StringUtils.equalsIgnoreCase(style, STYLE_PLAIN)) {
            return STYLE_CLASS_PLAIN;
        } else if (StringUtils.equalsIgnoreCase(style, STYLE_LIST)) {
            return STYLE_CLASS_LIST;
        } else {// default is menue
            return STYLE_CLASS_MENU;
        }
    }

    /**
     * Prepares the context map for non tabbed list views.
     * 
     * @param sort
     *            tells how the spaces shall be sorted
     * @param contextMap
     *            the context map to prepare
     * @param permittedSpaces
     */
    private void prepareDefaultContextMap(String sort, Map<String, Object> contextMap, int hierarchyDepth,
            Space parentSpace, List permittedSpaces) {

        Builder newQuery = SpacesQuery.newQuery();
        newQuery.withSpaceType(SpaceType.GLOBAL);
        HierarchyBuilder hierarchyBuilder = new HierarchyBuilder(ComponentManager.getSpaceManager().getAllSpaces(
                newQuery.build()), sort);

        hierarchyBuilder.setHierarchyDepth(hierarchyDepth);
        hierarchyBuilder.setParentSpace(parentSpace);
        hierarchyBuilder.setShowRoot((Boolean) contextMap.get(PARAMETER_SHOWROOT));
        hierarchyBuilder.init(permittedSpaces);
        contextMap.put("hierarchyBuilder", hierarchyBuilder);
    }

    /**
     * Prepares the contextMap for the tabbed list view (Confluence default).
     * <p>
     * This contains mainly code from the original getHtml().
     * </p>
     * 
     * @param displaySingleTab
     *            if only one tab should be displayed, this member contains the tab
     * @param sort
     *            tells how the spaces shall be sorted
     * @param contextMap
     *            the context map to prepare
     * @param permittedSpaces
     * @param hierarchyDepth
     */
    private void prepareListContextMap(String displaySingleTab, String sort, Map<String, Object> contextMap,
            int hierarchyDepth, Space parentSpace, List permittedSpaces) {

        DashboardMacroSupport dashboardSupport = getDashboardMacroSupport();
        if (TextUtils.stringSet(displaySingleTab)
                && Arrays.asList("my", "all", "new", "team").contains(displaySingleTab)) {
            contextMap.put("displaySingleTab", displaySingleTab);
            contextMap.put("spacesSelectedTab", displaySingleTab); // manually set the tab
        } else {
            contextMap.put("spacesSelectedTab", dashboardSupport.getSpacesSelectedTab());
        }

        List<String> availableTeams = new ArrayList<String>(new HashSet<String>(
                dashboardSupport.getViewableTeamLabels()));
        String selectedTeam = getSelectedTeam(availableTeams, dashboardSupport);

        // In shared mode, this method gets all viewable spaces and derives space group list for
        // this. This should better optimized when space groups get permissions.
        List<SpaceGroup> spaceGroups = dashboardSupport.getSpaceGroups();

        contextMap.put("viewableSpaceGroups", spaceGroups);
        contextMap.put("viewableTeamLabels", availableTeams);
        contextMap.put("selectedTeam", selectedTeam);

        // when this macro is called out with an HTTP request (such as a space export) then it
        // doesn't really matter that there is no baseurl.
        if (dashboardSupport.getRequest() != null) {
            contextMap.put(
                    "baseurl",
                    buildBaseUrl(dashboardSupport.getRequest(),
                            Arrays.asList("spacesSelectedTab", "selectedTeam", "selectedGroup")));
        }

        contextMap.put("personalSpace", spaceManager.getPersonalSpace(AuthenticatedUserThreadLocal.getUser()));
        contextMap.put("favouriteSpaces", dashboardSupport.getFavouriteSpaces());
        contextMap.put("recentlyCreatedSpaces", dashboardSupport.getNewSpaces());

        List<Space> spacesEditableByUser = spaceManager.getAllSpaces(SpacesQuery.newQuery()
                .forUser(AuthenticatedUserThreadLocal.getUser())
                .withPermission(Permission.EDIT.toString()).build());
        contextMap.put("spacesEditableByUser", spacesEditableByUser);

        // pull data from the db only if the relevant tab has been selected
        String tab = (String) contextMap.get("spacesSelectedTab");
        if ("all".equals(tab)) {

            SpacesQuery.Builder spaceQueryBuilder = SpacesQuery.newQuery();
            spaceQueryBuilder.withSpaceType(SpaceType.GLOBAL);
            List allSpaces = spaceManager.getAllSpaces(spaceQueryBuilder.build());
            HierarchyBuilder hierarchyBuilder = new HierarchyBuilder(allSpaces, sort);

            hierarchyBuilder.setHierarchyDepth(hierarchyDepth);
            hierarchyBuilder.setParentSpace(parentSpace);
            hierarchyBuilder.setShowRoot((Boolean) contextMap.get(PARAMETER_SHOWROOT));
            hierarchyBuilder.init(permittedSpaces);
            contextMap.put("hierarchyBuilder", hierarchyBuilder);
        }
        if ("groups".equals(tab)) {
            String selectedGroup = getSelectedGroup(spaceGroups, dashboardSupport);
            contextMap.put("selectedGroup", selectedGroup);
            contextMap.put("spacesForGroup", getSpacesForGroup(selectedGroup));
        }
        if ("team".equals(contextMap.get("spacesSelectedTab"))) {
            contextMap.put("spacesForTeam", getSpacesForTeam(selectedTeam));
        }

        if ("new".equals(contextMap.get("spacesSelectedTab"))) {
            contextMap.put("sortedRecentlyCreatedSpaces", getSortedRecentlyCreatedSpaces(dashboardSupport));
        }
    }

    private String buildBaseUrl(HttpServletRequest request, List ignoredKeys) {
        if (request == null) {
            return null;
        }

        String baseurl = GeneralUtil.appendAmpsandOrQuestionMark(request.getRequestURI());

        // put all previous request parameters back on
        // (except "selectedTeam" and "spacesSelectedTab" which may be there because we pass them
        // via url instead of posting)
        Enumeration keys = request.getParameterNames();
        while (keys.hasMoreElements()) {
            String key = (String) keys.nextElement();
            if (!ignoredKeys.contains(key)) {
                baseurl += GeneralUtil.urlEncode(key) + "=" + GeneralUtil.urlEncode(request.getParameter(key)) + "&";
            }
        }

        return baseurl;
    }

    /**
     * 
     * @return the Bandana manager
     */
    public BandanaManager getBandanaManager() {
        return bandanaManager;
    }

    /**
     * 
     * @param bandanaManager
     */
    public void setBandanaManager(BandanaManager bandanaManager) {
        this.bandanaManager = bandanaManager;
    }

    /**
     * Load data for the given space and key.
     * 
     * @param key
     *            the data key
     * @param context
     *            the space the data is stored for, if <code>null</code> the data will loaded from
     *            the global context
     * @return the data for the content entity
     */
    public Object loadViaBandana(Space context, String key) {
        return bandanaManager.getValue(new ConfluenceBandanaContext(context), key);
    }

    // ---------------------------------------------------------------------------------- //
    // The following code is taken from Confluence Dashboard Macros and condensed or changed where
    // necessary (Changes have been commented)

    protected DashboardMacroSupport getDashboardMacroSupport() {
        dashboardMacroSupport = createDashboardMacroSupport();
        dashboardMacroSupport.setLabelManager(labelManager);
        dashboardMacroSupport.setSpaceManager(spaceManager);
        dashboardMacroSupport.setSettingsManager(settingsManager);
        dashboardMacroSupport.setSpaceGroupManager(spaceGroupManager);
        dashboardMacroSupport.setLocaleManager(localeManager);
        return dashboardMacroSupport;
    }

    /**
     * Instantiates DashboardMacroSupport. This method is here because it is overridden in tests to
     * instead provide a mock DashboardMacroSupport.
     * 
     * @return A DashboradMacroSupport object.
     */
    protected DashboardMacroSupport createDashboardMacroSupport() {
        return new DashboardMacroSupport();
    }

    private Map<String, Object> getDefaultContextMap(Map macroParameter, RenderContext renderContext) {
        Map<String, Object> contextMap = getMacroVelocityContext();

        contextMap.put("i18NBean", getI18n());
        contextMap.put("labelManager", labelManager);

        // set the show space icons only for style menu
        contextMap.put(PARAM_SHOW_SPACE_ICONS, new Boolean((String) macroParameter.get(PARAM_SHOW_SPACE_ICONS))
                && (STYLE_MENU.equals(macroParameter.get(STYLE)) || macroParameter.get(STYLE) == null));
        // set the homebutton flag only if in menu mode
        contextMap.put(PARAM_SHOW_HOME_BUTTON, new Boolean((String) macroParameter.get(PARAM_SHOW_HOME_BUTTON))
                && (STYLE_MENU.equals(macroParameter.get(STYLE)) || macroParameter.get(STYLE) == null));

        String currentSpace = (String) macroParameter.get(PARAMETER_CURRENTSPACE);
        if (currentSpace == null) {
            PageContext context = (PageContext) renderContext;
            if (context != null && context.getEntity() != null) {
                long pageId = context.getEntity().getId();
                currentSpace = ComponentManager.getSpaceManager().getSpaceFromPageId(pageId);
            }
        }
        contextMap.put(PARAMETER_CURRENTSPACE, currentSpace);

        contextMap.put(PARAMETER_SHOWROOT, new Boolean((String) macroParameter.get(PARAMETER_SHOWROOT)));

        // look for width parameter
        String widthParameter = GeneralUtil.htmlEncode((String) macroParameter.get("width"));
        if (!TextUtils.stringSet(widthParameter)) {
            widthParameter = "100%";
        }
        contextMap.put("tableWidth", widthParameter);

        contextMap.put("remoteUser", AuthenticatedUserThreadLocal.getUser());
        contextMap.put("domainName", getGlobalSettings().getBaseUrl());

        // the following code handles additional parameters for sub spaces and was added
        String key = (String) macroParameter.get(KEY);
        if (StringUtils.isNotBlank(key) && ComponentManager.getSpaceManager().getSpace(key) == null) {
            throw new IllegalArgumentException(GeneralUtil.getI18n().getText("display.subspace.space.null"));
        }
        contextMap.put(KEY, key);

        String displayEmptyMessage = (String) macroParameter.get(DISPLAY_EMPTY_MESSAGE);
        if (displayEmptyMessage != null && displayEmptyMessage.equalsIgnoreCase("false")) {
            contextMap.put(DISPLAY_EMPTY_MESSAGE, false);
        } else {
            contextMap.put(DISPLAY_EMPTY_MESSAGE, true);
        }

        String addlink = (String) macroParameter.get(ADDLINK);
        List<Link> linkTree = new ArrayList<Link>();
        if (addlink != null) {
            if (addlink.equalsIgnoreCase("true")) {
                linkTree = linkTreeService.getLinkTree();
            }
        }
        contextMap.put("addlink", addlink);
        contextMap.put("linkTree", linkTree);

        return contextMap;
    }

    private Settings getGlobalSettings() {
        return settingsManager.getGlobalSettings();
    }

    private I18NBean getI18n() {
        return GeneralUtil.getI18n();
    }

    private Map<String, Object> getMacroVelocityContext() {
        return MacroUtils.defaultVelocityContext();
    }

    public void setSpaceGroupManager(SpaceGroupManager spaceGroupManager) {
        this.spaceGroupManager = spaceGroupManager;
    }

    private Set<Space> getPermittedSpaces() {
        return getDashboardMacroSupport().getPermittedSpaces();
    }

    public String getSelectedTeam(List<String> availableTeams, DashboardMacroSupport support) {
        if (availableTeams == null || availableTeams.size() == 0) {
            return "";
        }

        String result;

        String requestSelectedTeam = support.getRequestParameter("selectedTeam");
        if (requestSelectedTeam != null && availableTeams.contains(requestSelectedTeam)) {
            getUserInterfaceState().setDashboardSpacesSelectedTeam(requestSelectedTeam);
            result = support.getRequestParameter("selectedTeam");
        } else {
            String existingPreference = getUserInterfaceState().getDashboardSpacesSelectedTeam();
            if (TextUtils.stringSet(existingPreference) && availableTeams.contains(existingPreference)) {
                result = existingPreference;
            } else {
                result = availableTeams.get(0);
            }
        }

        return result;
    }

    public String getSelectedGroup(List<SpaceGroup> availableGroups, DashboardMacroSupport support) {
        if (availableGroups == null || availableGroups.size() == 0) {
            return "";
        }

        String result;

        String requestSelectedGroup = support.getRequestParameter("selectedGroup");
        if (requestSelectedGroup != null && doesSpaceGroupListContainKey(availableGroups, requestSelectedGroup)) {
            getUserInterfaceState().setDashboardSpacesSelectedGroup(requestSelectedGroup);
            result = support.getRequestParameter("selectedGroup");
        } else {
            String existingPreference = getUserInterfaceState().getDashboardSpacesSelectedGroup();
            if (TextUtils.stringSet(existingPreference)
                    && doesSpaceGroupListContainKey(availableGroups, existingPreference)) {
                result = existingPreference;
            } else {
                result = availableGroups.get(0).getKey();
            }
        }

        return result;
    }

    private boolean doesSpaceGroupListContainKey(List<SpaceGroup> spaceGroups, String key) {
        for (SpaceGroup spaceGroup : spaceGroups) {
            if (spaceGroup.getKey().equals(key)) {
                return true;
            }
        }
        return false;
    }

    public UserInterfaceState getUserInterfaceState() {
        return new UserInterfaceState(AuthenticatedUserThreadLocal.getUser());
    }

    public List<Space> getSpacesForGroup(String selectedGroup) {
        if (TextUtils.stringSet(selectedGroup)) {
            SpaceGroup group = spaceGroupManager.getSpaceGroup(selectedGroup);
            if (group != null) {
                User user = AuthenticatedUserThreadLocal.getUser();
                List<Space> spaces = spaceManager.getAllSpaces(SpacesQuery.newQuery().forUser(user)
                        .inSpaceGroup(group).build());
                Collections.sort(spaces, new SpaceComparator());
                return spaces;
            } else {
                getUserInterfaceState().setDashboardSpacesSelectedGroup(null); // reset this
                // preference if
                // group referred to
                // here no longer
                // exists
            }
        }

        return Collections.emptyList();
    }

    public List<Space> getSpacesForTeam(String selectedTeam) {
        Set<Space> permittedSpaces = getPermittedSpaces();
        if (TextUtils.stringSet(selectedTeam)) {
            Label label = labelManager.getLabel("team:" + selectedTeam);
            if (label != null) {
                List<Space> spaces = labelManager.getSpacesWithLabel(label);
                spaces.retainAll(permittedSpaces);
                Collections.sort(spaces, new SpaceComparator());
                return spaces;
            } else {
                getUserInterfaceState().setDashboardSpacesSelectedTeam(null); // reset this
                // preference if label
                // referred to here no
                // longer exists
            }
        }

        return Collections.emptyList();
    }

    public Map<String, List<Space>> getSortedRecentlyCreatedSpaces(DashboardMacroSupport support) {
        Map<AbstractInstant, List<Space>> datesToSpaceLists = new TreeMap<AbstractInstant, List<Space>>(
                new ReverseComparator());

        for (Space space : support.getNewSpaces()) {
            // Remove any time stored by creation date (so that we can group by date only)
            final AbstractInstant key = new DateMidnight(space.getCreationDate().getTime());

            if (datesToSpaceLists.get(key) == null) {
                List<Space> tempSpaceList = new ArrayList<Space>();
                tempSpaceList.add(space);
                datesToSpaceLists.put(key, tempSpaceList);
            } else {
                datesToSpaceLists.get(key).add(space);
            }
        }

        Map<String, List<Space>> result = new LinkedHashMap<String, List<Space>>();
        for (Map.Entry<AbstractInstant, List<Space>> entry : datesToSpaceLists.entrySet()) {
            result.put(support.getDateFormatter().formatDateFull(entry.getKey().toDate()), entry.getValue());
        }

        // sort the individual lists.
        for (List<Space> spaces : result.values()) {
            Collections.sort(spaces, new SpaceComparator());
        }

        return result;
    }

    public void setSpaceManager(SpaceManager spaceManager) {
        this.spaceManager = spaceManager;
    }

    public void setLinkTreeService(LinkTreeService linkTreeService) {
        this.linkTreeService = linkTreeService;
    }

    public void setLabelManager(LabelManager labelManager) {
        this.labelManager = labelManager;
    }

    public void setSettingsManager(SettingsManager settingsManager) {
        this.settingsManager = settingsManager;
    }

    public void setLocaleManager(final LocaleManager localeManager) {
        this.localeManager = localeManager;
    }

    public RenderMode getBodyRenderMode() {
        return RenderMode.NO_RENDER;
    }

    public boolean hasBody() {
        return false;
    }
}
